from delphai_search_utils.utils import remove_conjunction, remove_stopwords
from delphai_search_utils.utils import is_masked, remove_mask, remove_elements_from_text
from delphai_search_utils.utils import split_with_indices
import pytest
import os
from typing import List, Dict, Tuple
os.environ['DELPHAI_ENVIRONMENT'] = 'development'


@pytest.mark.parametrize("query,output", [
  ("urban mobility in", "urban mobility __"),
  ("robots and cars after", "robots and cars _____"),
  ("robots and cars in germany", "robots and cars in germany")
])
def test_conjunction_removal(query: str, output: str):
  assert remove_conjunction(query) == output
  

@pytest.mark.parametrize("query,output", [
  ("urban mobility companies", "urban mobility _________"),
  ("urban mobility companies in Prague", "urban mobility _________ in Prague"),
  ("urban mobility founded after 2020", "urban mobility _______ after 2020"),
])
def test_remove_stopwords(query: str, output: str):
  assert remove_stopwords(query) == output


@pytest.mark.parametrize("query,output", [
  ("____", True),
  ("urban", False),
  ("__ban", False),
])
def test_is_masked(query: str, output: bool):
  assert is_masked(query) == output


@pytest.mark.parametrize("query,output", [
  ("urban mobility ________", "urban mobility"),
  ("mobility ________ _____ after 2020", "mobility after 2020"),
  ("urban mobility", "urban mobility")
])
def test_remove_mask(query: str, output: str):
  assert remove_mask(query) == output


@pytest.mark.parametrize("query,labels,output", [
  ('urban mobility in germany', [{'start_index': 18, 'end_index': 25}], 'urban mobility __ _______'),
  ('german cars', [{'start_index': 0, 'end_index': 6}], '______ cars'),
])
def test_remove_elements(query: str, labels: List[Dict], output: str):
  assert remove_elements_from_text(query, labels, remove_conjunctions=True) == output


@pytest.mark.parametrize("query,output", [
  ('urban mobility in germany', [('urban', 0), ('mobility', 6), ('in', 15), ('germany', 18)]),
  ('electric vehicles', [('electric', 0), ('vehicles', 9)]),
  ('robots and cats', [('robots', 0), ('and', 7), ('cats', 11)]),
])
def test_index_splitting(query: str, output: List[Tuple[int, int]]):
  tokens = list(split_with_indices(query))
  assert len(tokens) == len(query.split())
  assert all(token in query for token, _ in tokens)
  assert all(' ' not in token for token, _ in tokens)
  assert all(query[start_index:].startswith(token) for token, start_index in tokens)



""" semantic processing of the query to generate better results."""

from __future__ import annotations
from collections import namedtuple, defaultdict
import logging
from typing import List, Callable, Dict
from delphai_search_utils.utils import split_with_indices

# which keywords to replace with AND?
AND_KEYWORDS = ["and", "for"]
OR_KEYWORDS = ["or"]
NOT_KEYWORDS = ["not"]

SYNONYM_TERMS_TYPE = "synonyms"
SIMILAR_TERMS_TYPE = "similars"

# This represents a related term, e.g. a synonym or an alternative spelling,
# but it can also be a boolean query. The .query field is saved as a TextQuery,
# so it will be transformed to a string again afterwards.
RelatedQueryTerm = namedtuple('RelatedQueryTerm', ['query', 'source'])


class TextQueryOp:

    def __init__(self, op):
        if op not in ['OR', 'AND', 'NOT']:
            logging.error(f'TextQueryOp has been called with an invalid argument: {op}')
        self.op = op

    def negate(self):
        if self.op == 'OR':
            self.op = 'AND'
        elif self.op == 'AND':
            self.op = 'OR'
        elif self.op == 'NOT':
            # This case happens in the current logic but it's not resolved
            # here so it can be just ignored.
            pass


class TextQueryTerm:

    def __init__(
            self,
            term: str,    # actual term string
            start_pos: int = None,    # character position
            negated: bool = False,    # whether the term is negated
            quoted: bool = False,    # whether the term is quoted
            related_query_terms: Dict[int, List[RelatedQueryTerm]] = None,    # list of related terms
    ):
        self.str = term
        self.start_pos = start_pos
        self.negated = negated
        self.quoted = term.startswith('"') and term.endswith('"')
        self.related_query_terms = defaultdict(list)
        if related_query_terms is not None:
            self.related_query_terms = related_query_terms

    def expand_with_fct(
            self,
            similars_fct: Callable,
            terms_type: int = SYNONYM_TERMS_TYPE,    # or SIMILAR_TERMS_TYPE
            apply_on_synonyms: bool = True,
    ):
        """ Adds related terms to this term, depending on the function that
        is passed.
        'apply_on_synonyms' determines if we also apply the function on the
        synonyms that were already found or only on the term itself."""
        if self.quoted:
            return
        apply_on = [self.str]
        if apply_on_synonyms:
            # collect related query terms that only have a single term
            # (so they are e.g. not queries with boolean operators).
            terms_collection = [
                query_term.query.get_terms()
                for query_term in self.related_query_terms[SYNONYM_TERMS_TYPE]]
            apply_on += [
                terms[0].str for terms in terms_collection
                if len(terms) == 1 and not terms[0].quoted]
        for term in apply_on:
            similars_result = similars_fct(term)
            for result in similars_result:
                self.add_related(terms_type, result, similars_fct.__name__)

    def add_related(self, term_type: str, related_query_term: str, source: str):
        query_object = TextQuery.from_str(related_query_term)
        related_object = RelatedQueryTerm(query_object, source)
        self.related_query_terms[term_type].append(related_object)

    def negate(self):
        self.negated = not self.negated

    def __str__(self):
        # 'add_quotes_to_unquoted_terms' is a parameter to influence the Elasticsearch behavior;
        # if we change this to False the term will not be treated as a single phrase but
        # implicitly combined with an AND, so there will be a lot more (irrelevant) results.
        add_quotes_to_unquoted_terms = True
        self_term = self.str
        if not self.quoted and add_quotes_to_unquoted_terms:
            self_term = f'"{self.str}"'
        terms = [self_term] + \
            [term.query.to_str() for term in self.related_query_terms[SYNONYM_TERMS_TYPE]]
        terms_str = ' OR '.join(terms)
        if len(terms) >= 2:
            terms_str = f'({terms_str})'
        if self.negated:
            terms_str = f'NOT {terms_str}'
        return terms_str

    def __repr__(self):
        return self.__str__()


class TextQuery:
    """ helper class to help with query expansion etc.
    a TextQuery consists of terms, operators, and subqueries (of type TextQuery).
    currently supports AND, OR and NOT."""

    def __init__(self, subqueries, prune=False):
        self.subqueries = subqueries
        if prune:
            self.substitute_negations()
            self.collapse()

    def to_str(self):
        """ returns a readable query string representation which
        can be used in Elasticsearch."""
        query_parts = []
        for i, subq in enumerate(self.subqueries):
            if type(subq) == TextQuery:
                if len(subq.subqueries) == 1:
                    query_parts.append(subq.to_str())
                else:
                    query_parts += ['(', subq.to_str(), ')']
            elif type(subq) == TextQueryTerm:
                query_parts.append(str(subq))
            elif type(subq) == TextQueryOp:
                query_parts.append(subq.op)
        query_str = " ".join(query_parts)
        if len(self.subqueries) >= 2:
            query_str = f'({query_str})'
        return query_str

    @classmethod
    def from_str(cls, query_str):
        """ query parser that identifies brackets, AND and OR.
        handles an arbitrary number of nested brackets but won't produce
        reasonable results for a wrong query syntax."""
        query_layers = defaultdict(list)
        current_layer = 0
        # create a stack and process tokens from front to back
        query_stack = list(split_with_indices(query_str))[::-1]
        open_quotes = False
        while len(query_stack) >= 1:
            tok, start_idx = query_stack.pop()
            # preprocess token
            # tok = tok.lower()
            tok_lower = tok.lower()
            if tok == '"':
                open_quotes = not open_quotes
            if len(tok) == 0:
                continue
            # handle quotes
            if '"' in tok and not tok == '"':
                # split by quotes
                quote_idx = tok.index('"')
                if quote_idx + 1 < len(tok):
                    query_stack.append((tok[quote_idx + 1:], start_idx + quote_idx + 1))
                query_stack.append(('"', start_idx + quote_idx))
                if quote_idx > 0:
                    query_stack.append((tok[:quote_idx], start_idx))
            # handle brackets
            elif tok.startswith("(") and not open_quotes:
                current_layer += 1
                remaining = tok[1:]
                query_stack.append((remaining, start_idx + 1))
            elif tok.endswith(")") and not tok == ')' and not open_quotes:
                query_stack.append((')', start_idx + len(tok) - 1))
                query_stack.append((tok[:-1], start_idx))
            elif tok == ')' and not open_quotes:
                nested_query = TextQuery.from_list(query_layers[current_layer])
                query_layers[current_layer - 1].append(nested_query)
                query_layers[current_layer] = []
                current_layer -= 1
            # handle actual terms
            else:
                # keywords
                if tok_lower in AND_KEYWORDS and not open_quotes:
                    query_layers[current_layer].append(TextQueryOp('AND'))
                elif tok_lower in OR_KEYWORDS and not open_quotes:
                    query_layers[current_layer].append(TextQueryOp('OR'))
                elif tok_lower in NOT_KEYWORDS and not open_quotes:
                    query_layers[current_layer].append(TextQueryOp('NOT'))
                # query terms
                elif len(query_layers[current_layer]) == 0:
                    query_layers[current_layer].append(TextQueryTerm(tok, start_idx))
                elif type(query_layers[current_layer][-1]) in (TextQueryOp, TextQuery):
                    # new term
                    query_layers[current_layer].append(TextQueryTerm(tok, start_idx))
                elif type(query_layers[current_layer][-1]) == TextQueryTerm:
                    # append to current term
                    term = query_layers[current_layer][-1]
                    end_pos = start_idx + len(tok)
                    # print(f'appending {tok} to {term}')
                    new_term = query_str[term.start_pos:end_pos]
                    query_layers[current_layer][-1] = TextQueryTerm(new_term, term.start_pos)
        return cls(query_layers[0], prune=True)

    @classmethod
    def from_list(cls, query_parts):
        """ create query from list or tuple of query parts."""
        subquery_list = list(query_parts)
        return cls(subquery_list)

    def get_terms(self) -> List[TextQueryTerm]:
        """ returns only the actual terms included in this query without
        any operators; is used for resolving keywords."""
        terms = []
        for subq in self.subqueries:
            if type(subq) == TextQueryTerm:
                terms.append(subq)
            elif type(subq) == TextQuery:
                terms += subq.get_terms()
        return terms

    def reduce(self):
        """ reduces this query if it has only a single subquery"""
        if len(self.subqueries) == 1:
            subq = self.subqueries[0]
            if type(subq) == TextQuery:
                return subq.reduce()
            else:
                return [subq]
        else:
            return self.subqueries

    def collapse(self):
        """ optimize the query: if it has only one TextQuery as subquery
        then one layer can be removed."""
        collapsed_qs = []
        subs = self.reduce()
        for i, subq in enumerate(subs):
            if type(subq) == TextQuery:
                if not subq.is_empty():
                    # collapse subquery and append
                    subq.collapse()
                    collapsed_qs.append(subq)
            elif type(subq) in [TextQueryTerm, TextQueryOp]:
                # regular term
                collapsed_qs.append(subq)
        self.subqueries = collapsed_qs

    def negate(self):
        for subquery in self.subqueries:
            subquery.negate()

    def substitute_negations(self):
        """ Substitute the query so that negations are passed on to
        the single query terms; e.g.
        'NOT (A AND B)' becomes '(NOT A OR NOT B)'.
        see also https://en.wikipedia.org/wiki/De_Morgan%27s_laws
        """
        # Identify all query parts that need to be negated in the right order.
        # We want to substitute nested negations first because that makes
        # it very straight forward. E.g.
        # 'NOT (A AND NOT (B OR C))'
        # will be 'NOT (A AND (NOT B AND NOT C))'
        # and then '(NOT A OR (B OR C))'
        def build_stack(query, stack):
            for i, subquery in enumerate(query.subqueries):
                if type(subquery) == TextQueryOp and subquery.op == 'NOT':
                    # next term should be negated
                    try:
                        stack.append(query.subqueries[i + 1])
                    except IndexError:
                        logging.error(f'Invalid query: Found NOT at the end of a query: {query.to_str()}')
                elif type(subquery) == TextQuery:
                    build_stack(subquery, stack)

        negate_stack = []
        build_stack(self, negate_stack)
        # Do the actual substitutions
        for subquery in negate_stack[::-1]:
            subquery.negate()
        # Remove all 'NOT's because they have been applied but still exist.
        # (They are now encoded in TextQueryTerm.negated.)
        query_list = [self]
        while len(query_list) >= 1:
            query = query_list.pop()
            query.subqueries = [
                subquery for subquery in query.subqueries
                if not (type(subquery) == TextQueryOp and subquery.op == 'NOT')
            ]
            query_list += [
                subquery for subquery in query.subqueries
                if type(subquery) == TextQuery
            ]

    def is_empty(self):
        """ returns true if it doesn't contain relevant query information"""
        if len(self.subqueries) == 0:
            return True
        else:
            for subq in self.subqueries:
                if type(subq) == TextQueryTerm or    \
                        (type(subq) == TextQuery and not subq.is_empty()):
                    return False
            return True

    def expand_with_syns_fct(
            self,
            similars_fct: Callable,
            terms_type: int = SYNONYM_TERMS_TYPE,    # or SIMILAR_TERMS_TYPE
            apply_on_synonyms: bool = True):
        for query_term in self.get_terms():
            query_term.expand_with_fct(similars_fct, terms_type, apply_on_synonyms)

    def get_related_terms_map(self) -> Dict[str, List[Dict[str, str]]]:
        """ Create a map of related terms by collecting them for each unique term."""
        related_terms_map = defaultdict(list)
        processed_terms = set()
        for query_term in self.get_terms():
            if query_term.str in processed_terms:
                # avoid duplicates in the map;
                # equal terms will have equal related terms.
                continue
            processed_terms.add(query_term.str)
            for related_type, terms in query_term.related_query_terms.items():
                for related_term in terms:
                    related_terms_map[related_type].append({
                        'original': query_term.str,
                        'term': related_term.query.to_str(),
                        'source': related_term.source,
                    })
        return related_terms_map

    def add_related_terms_from_map(self, related_terms_map: Dict[str, List[Dict[str, str]]]):
        """ Takes a map of related terms as created by get_related_terms_map()
        and adds them to the terms of this query."""
        for query_term in self.get_terms():
            for related_type, terms_list in related_terms_map.items():
                for related_term in terms_list:
                    if related_term['original'] == query_term.str:
                        related_term_str = str(related_term['term'])
                        source = related_term.get('source', '')
                        query_term.add_related(
                            related_type,
                            related_term_str,
                            source)
                        if related_type == 'synonyms':
                            print(f'added {related_term_str=} to {query_term=}')

    def __str__(self):
        return self.to_str()

    def __repr__(self):
        return self.__str__()



from delphai_search_utils.text_query import TextQuery
from delphai_utils.config import get_config
from delphai_utils.formatting import clean_url
from collections import defaultdict


def get_field_weights():
    weights_cfg = get_config('elastic_ranking.query_weights')
    labels_cfg = get_config('label_layers_searched')
    for label, incl in labels_cfg.items():
        label = f'{label}_labels'
        if label in weights_cfg and not incl:
            weights_cfg.pop(label)
        if label not in weights_cfg and incl:
            weights_cfg[label] = 1
    return weights_cfg


def build_query_clause(term, weight_factor=1, with_quotes=True, incl_names=False):
    """ simply builds a query clause and adjusts the weights"""
    weights_cfg = get_field_weights()
    incl_text = not get_config('elastic_ranking.disable_webpage_search')
    fields = [
        f'{name}^{weight / weight_factor:.3f}'
        for name, weight in weights_cfg.items()
        if (incl_names or name not in ['name', 'url', 'name.keyword', 'url.keyword']) and
            (incl_text or name not in ['text', 'pages.text'])]
    if with_quotes:
        unquoted_term = term.strip('" ')
        term = f'"{unquoted_term}"'
    # build elastic clause
    clause = {
        "query_string": {
            "query": term,
            "fields": fields,
            "type": 'best_fields',  # default
        }
    }
    if get_config('elastic_ranking.sum_field_scores', False):
        # alternative scoring which sums score of each field for a term,
        # instead of taking the maximum score
        clause['query_string']['type'] = 'cross_fields'
        clause['query_string']['tie_breaker'] = 1.0
    return clause


def create_elastic_query(query_parsed, related_terms_map):
    query_expanded = TextQuery.from_str(query_parsed)
    query_expanded.add_related_terms_from_map(related_terms_map)
    query_expanded_str = query_expanded.to_str()
    query_dictionary = {
        "must": [],
        "should": [],
        "string": query_expanded_str,
    }
    top_query = {
        "bool": {
            "minimum_should_match": 1,
            "should": [
                # query part with whole query string
                build_query_clause(query_expanded_str, with_quotes=False, incl_names=True),
                # exact match for name
                {
                    "term": {
                        "name_keyword": {
                            "term": query_parsed.strip('"'),
                            "boost": 2
                        }
                    }
                },
                # exact match for url
                {
                    "term": {
                        "url.keyword": {
                            "term": clean_url(query_parsed),
                            "boost": 3
                        }
                    }
                }
            ]
        }
    }
    if 'project_annotations' in get_config('elastic_ranking.query_weights'):
        weight = get_config('elastic_ranking.query_weights.project_annotations')
        anno_match_clause = {
            "term": {
                "project_annotations": {
                    "value": query_parsed,
                    "boost": weight * 2,
                }
            }
        }
        top_query['bool']['should'].append(anno_match_clause)
    query_dictionary['must'] = top_query

    # add single terms as additional boosts 
    # (with weights proportional to their position in the query)
    term_nbr = len(query_expanded.get_terms())
    # The factors are inversely proportional to term weights
    factors = []
    if term_nbr < 2:
        # If only one term is present, it's factor is set to 2
        factors = [2 for i in range(term_nbr)]
    else:
        # Otherwise, the factors range from 1.3 to 3.9
        factors = [1.3 + 2.6/(term_nbr - 1)*i for i in range(term_nbr)]
    for factor, query_term in list(zip(factors, query_expanded.get_terms())):
        if not query_term.negated:
            # add query boosting query clause for this similar term;
            # we don't need quotes here because they are already added in the term's
            # __str__() function itself; for complex queries this leads to errors, e.g.
            # '"("EVs" OR "electric vehicles")"', so that the brackets are enclosed in quotes
            term_clause = build_query_clause(str(query_term), factor, with_quotes=False, incl_names=False)
            query_dictionary['should'].append(term_clause)

    # add related similar terms as additional boosts
    # (with weights proportional to the position of the original term in the query)
    # (max N per query term)
    max_term_count = get_config('elastic_ranking.similar_keywords.count')
    per_term_counts = defaultdict(int)
    # If a related term is not matched with any original terms, it's factor is set to 4
    factors.append(4)
    for similar_term in related_terms_map['similar_terms']:
        if per_term_counts[similar_term['original']] < max_term_count:
            ind = 0
            for query_term in query_expanded.get_terms():
                # Checking which query term corresponds to this similar term
                if query_term.str == similar_term['original']:
                    break
                ind += 1
            factor = factors[ind]
            term_clause = build_query_clause(similar_term['term'], factor, with_quotes=True, incl_names=False)
            query_dictionary['should'].append(term_clause)
            per_term_counts[similar_term['original']] += 1

    query_dictionary['_source'] = {
        'include': [
            'see_also_ids',
            'company_id',
            'name',
            'url',
            'href',
            'organisation_type',
            'founded',
            'headquarters',
            'relevance_score',
            'description',
            'funding',
            'n_investors',
            'added'
        ]
    }
    return query_dictionary



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['delphai_search_utils', 'delphai_search_utils.tests']

package_data = \
{'': ['*'], 'delphai_search_utils': ['data/*']}

install_requires = \
['delphai-utils[config,grpc,database]>=0.2.20,<0.3.0', 'nltk', 'sacremoses']

setup_kwargs = {
    'name': 'delphai-search-utils',
    'version': '0.1.2',
    'description': 'delphai search utilities',
    'long_description': None,
    'author': 'Christopher Schiefer',
    'author_email': 'christopher.schiefer@delphai.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/delphai/delphai-search-utils',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from sacremoses import MosesTruecaser
from os.path import join
from nltk.stem.snowball import SnowballStemmer
import logging
import nltk
from copy import deepcopy
from itertools import groupby
import json
import requests
import bson
from typing import Tuple, List, Generator


data_path = './delphai_search_utils/data/'
mtr = MosesTruecaser(join(data_path, 'demonyms.truecasemodel'))

special_stopwords = ['compani', 'found',]
stemmer = SnowballStemmer(language='english')


def remove_conjunction(query_str: str) -> str:
    """ Removes conjunctions at the end of the query which are very likely
    not related to the actual query terms but just left overs from filters etc."""
    if len(query_str.strip()) == 0:
        return query_str
    pos_tags = nltk.pos_tag(nltk.word_tokenize(query_str))
    if not pos_tags:
        return query_str
    # iterate through pos tags and keep absolute string positions in mind
    for i, ((tok, pos), (_, pos_tag)) in enumerate(zip(split_with_indices(query_str), pos_tags)):
        if i == len(pos_tags) - 1 or is_masked(pos_tags[i + 1][0]):
            if pos_tag in ['CC', 'IN']:
                query_str = mask_substr(query_str, pos, pos + len(tok))
    return query_str


def remove_stopwords(query_str: str) -> str:
    # remove special_stopwords
    query_stops = [word for word in query_str.split() if stemmer.stem(word) in special_stopwords]
    for stop_word in query_stops:
        stop_index = query_str.index(stop_word)
        if stop_index >= 0:
            query_str = mask_substr(query_str, stop_index, stop_index + len(stop_word))
    return query_str


def is_masked(substring: str) -> bool:
    return (all(char == '_' for char in substring))


def remove_mask(query_str: str) -> str:
    """" Remove masked words;
    Doesn't guarantee that character indices stay the same afterwards,
    e.g. if masked words are removed in the middle of the query."""
    unmasked = query_str.replace('_', ' ')
    return ' '.join(unmasked.split())


def mask_substr(query_str: str, start_index: int, end_index: int) -> str:
    """ replaces the element between the given indices with underscores so that
    it is ignored for the subsequent filter detection; they are removed afterwards
    but kept for now to keep the correct character indices afterwards.
    This is not an ideal solution, so replacing it e.g. with a custom class which has
    a 'hide'-flag for each token would seem more resilient. (spaCy could do that..?)"""
    return query_str[:start_index] + '_' * (end_index - start_index) + query_str[end_index:]


def remove_elements_from_text(query_str: str, labels: List, remove_conjunctions=False) -> str:
    """ Removes the given labels from the query string and potentially also
    conjunctions; returns a cleaned query string."""
    for label in labels:
        query_str = mask_substr(query_str, label['start_index'], label['end_index'])
        if remove_conjunctions:
            query_str = remove_conjunction(query_str)
            query_str = remove_conjunction(query_str)
    return query_str


def add_label(labels, substring, label_type, start_ind, end_ind):
    labels.append({
        'start_index': start_ind,
        'end_index': end_ind,
        'substring': substring,
        'label_type': label_type
    })


def add_related_terms(elements, terms_dict, source, related_type):
    for original, terms in terms_dict.items():
        for term in terms:
            elements[related_type].append({
                'original': original,
                'term': term,
                'source': source,
            })


def add_to_log(changes_log, method, input_term, outputs):
    if changes_log is not None:
        # logging.info(f'adding {outputs} for {input_term} in {method}')
        changes_log[method].append({input_term: deepcopy(outputs)})


def truecase(query_str: str) -> str:
    """ (annoying) workaround to deal with multiple whitespaces etc."""
    truecased = query_str
    try:
        true_tokens = mtr.truecase(query_str, use_known=True)
        # returns list of tokens, re-assemble to proper string
        # with right amount of whitespaces etc.
        true_chars = ''.join(true_tokens)
        assembled = []
        true_index = 0
        for i, char in enumerate(query_str):
            if true_index < len(true_chars) and char.lower() == true_chars[true_index].lower():
                assembled.append(true_chars[true_index])
                true_index += 1
            else:
                assembled.append(char)
        truecased = ''.join(assembled)
    except Exception as exc:
        logging.warning(f'Exception in truecase() method: {exc}')
    return truecased


def api_wrap(api_info, request, max_timeouts=3, timeout_time=1.2):
    request_data = json.dumps(request)
    # Set the content type
    headers = {'Content-Type': 'application/json'}
    # If authentication is enabled, set the authorization header
    headers['Authorization'] = f'Bearer {api_info["key"]}'
    resp_raw = None
    while (resp_raw is None or resp_raw.text is None) and max_timeouts >= 0:
        try:
            resp_raw = requests.post(api_info["url"], request_data, headers=headers, timeout=timeout_time)
        except requests.exceptions.Timeout:
            logging.info(f'request to {api_url} timed out...')
        max_timeouts -= 1
    if resp_raw is None or resp_raw.text is None or resp_raw.status_code >= 400:
        return None
    else:
        return json.loads(resp_raw.text)


def is_object_id(term: str) -> bool:
    return bson.objectid.ObjectId.is_valid(term)


# https://stackoverflow.com/a/13734815
def split_with_indices(query: str, sep: str = ' ')\
        -> Generator[Tuple[str, int], None, None]:
    """ Splits the string by the given separator and returns all tokens
    together with their character start indices."""
    p = 0
    for k, g in groupby(query, lambda x: x == sep):
        q = p + sum(1 for i in g)
        if not k:
            yield query[p:q], p
        p = q
